# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import cairo

from typing import Callable
from gi.repository import Gtk, Gdk, Gio, Pango, PangoCairo, GdkPixbuf
from enum import Enum
import math
from gradia.backend.logger import Logger
from gradia.utils.colors import has_visible_color
import time
import unicodedata

logging = Logger()

start_time_seed = int(time.time())


class DrawingMode(Enum):
    SELECT = "SELECT"
    PEN = "PEN"
    TEXT = "TEXT"
    LINE = "LINE"
    ARROW = "ARROW"
    SQUARE = "SQUARE"
    CIRCLE = "CIRCLE"
    HIGHLIGHTER = "HIGHLIGHTER"
    CENSOR = "CENSOR"
    NUMBER = "NUMBER"

    def label(self):
        return {
            "PEN": _("Pen"),
            "ARROW": _("Arrow"),
            "LINE": _("Line"),
            "SQUARE": _("Rectangle"),
            "CIRCLE": _("Oval"),
            "TEXT": _("Text"),
            "SELECT": _("Select"),
            "HIGHLIGHTER": _("Highlighter"),
            "CENSOR": _("Censor"),
            "NUMBER": _("Number"),
        }[self.value]

    @property
    def shortcuts(self):
        return DrawingMode._shortcuts[self]

DrawingMode._shortcuts = {
    DrawingMode.SELECT:       ["0", "KP_0", "grave", "s"],
    DrawingMode.PEN:          ["1", "KP_1", "d", "p"],
    DrawingMode.TEXT:         ["2", "KP_2", "t"],
    DrawingMode.LINE:         ["3", "KP_3", "l"],
    DrawingMode.ARROW:        ["4", "KP_4", "a"],
    DrawingMode.SQUARE:       ["5", "KP_5", "r"],
    DrawingMode.CIRCLE:       ["6", "KP_6", "o"],
    DrawingMode.HIGHLIGHTER:  ["7", "KP_7", "h"],
    DrawingMode.CENSOR:       ["8", "KP_8", "c"],
    DrawingMode.NUMBER:       ["9", "KP_9", "n"],
}


class QuadBounds:
    def __init__(self, p1: tuple[float, float], p2: tuple[float, float], p3: tuple[float, float], p4: tuple[float, float]):
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.p4 = p4

    @classmethod
    def from_rect(cls, min_x: float, min_y: float, max_x: float, max_y: float):
        return cls(
            (min_x, min_y),
            (max_x, min_y),
            (max_x, max_y),
            (min_x, max_y)
        )

    @classmethod
    def from_start_end(cls, start: tuple[float, float], end: tuple[float, float]):
        min_x = min(start[0], end[0])
        max_x = max(start[0], end[0])
        min_y = min(start[1], end[1])
        max_y = max(start[1], end[1])
        return cls.from_rect(min_x, min_y, max_x, max_y)

    def get_points(self) -> list[tuple[float, float]]:
        return [self.p1, self.p2, self.p3, self.p4]

    def get_bounding_rect(self) -> tuple[float, float, float, float]:
        points = self.get_points()
        xs = [p[0] for p in points]
        ys = [p[1] for p in points]
        return min(xs), min(ys), max(xs), max(ys)


class DrawingAction:
    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        raise NotImplementedError

    def get_bounds(self) -> QuadBounds:
        raise NotImplementedError

    def contains_point(self, x_img: int, y_img: int) -> bool:
        min_x, min_y, max_x, max_y = self.get_bounds().get_bounding_rect()
        if isinstance(self, (LineAction, ArrowAction)):
            px, py = x_img, y_img
            x1, y1 = self.start
            x2, y2 = self.end
            line_len_sq = (x2 - x1) ** 2 + (y2 - y1) ** 2
            if line_len_sq == 0:
                return math.hypot(px - x1, py - y1) < (5 + self.options.size * 1.75)
            t = ((px - x1) * (x2 - x1) + (py - y1) * (y2 - y1)) / line_len_sq
            t = max(0, min(1, t))
            closest_x = x1 + t * (x2 - x1)
            closest_y = y1 + t * (y2 - y1)
            dist_sq = (px - closest_x)**2 + (py - closest_y)**2
            return dist_sq < (5 + self.options.size * 1.75)**2
        return min_x <= x_img <= max_x and min_y <= y_img <= max_y

    def _calculate_shadow_color(self, color):
        r = color.red
        g = color.green
        b = color.blue
        luminance = 0.299 * r + 0.587 * g + 0.114 * b
        if luminance < 0.5:
            return (1.0, 1.0, 1.0, 0.05)
        else:
            return (0.0, 0.0, 0.0, 0.3)

    def translate(self, dx: int, dy: int):
        raise NotImplementedError

    def get_drawing_mode(self) -> DrawingMode:
        return self.options.mode


class StrokeAction(DrawingAction):
    def __init__(self, stroke: list[tuple[int, int]], options):
        self.stroke = stroke
        self.options = options
        self._bounds = None

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        if len(self.stroke) < 2:
            return
        coords = [image_to_widget_coords(x, y) for x, y in self.stroke]
        line_width = self.options.size * scale
        self._build_path(cr, coords)
        cr.set_source_rgba(*self.options.primary_color)
        cr.set_line_width(line_width)
        cr.stroke()

    def _build_path(self, cr, coords):
        cr.set_line_cap(cairo.LineCap.ROUND)
        cr.set_line_join(cairo.LineJoin.ROUND)
        if len(coords) <= 2:
            cr.move_to(*coords[0])
            if len(coords) == 2:
                cr.line_to(*coords[1])
            return
        cr.move_to(*coords[0])
        for i in range(1, len(coords) - 1):
            x1, y1 = coords[i]
            x2, y2 = coords[i + 1]
            mid_x = (x1 + x2) * 0.5
            mid_y = (y1 + y2) * 0.5
            cr.curve_to(x1, y1, x1, y1, mid_x, mid_y)
        cr.line_to(*coords[-1])

    def get_bounds(self) -> QuadBounds:
        if self._bounds is None:
            if not self.stroke:
                self._bounds = QuadBounds.from_rect(0, 0, 0, 0)
            else:
                xs, ys = zip(*self.stroke)
                padding = self.options.size // 2
                self._bounds = QuadBounds.from_rect(
                    min(xs) - padding,
                    min(ys) - padding,
                    max(xs) + padding,
                    max(ys) + padding
                )
        return self._bounds

    def translate(self, dx: int, dy: int):
        self.stroke = [(x + dx, y + dy) for x, y in self.stroke]
        self._bounds = None


class ArrowAction(DrawingAction):
    ARROW_HEAD_SIZE_MULTIPLIER = 3
    MIN_DISTANCE_THRESHOLD = 2
    ARROW_ANGLE = math.pi / 6
    MIN_ARROW_SCALE = 0.3
    SCALE_DISTANCE_THRESHOLD = 120

    def __init__(self, start: tuple[int, int], end: tuple[int, int], shift: bool, options):
        self.options = options
        self.start = start
        if shift:
            dx = abs(end[0] - start[0])
            dy = abs(end[1] - start[1])
            if dx > dy:
                self.end = (end[0], start[1])
            else:
                self.end = (start[0], end[1])
        else:
            self.end = end

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        start_x, start_y = image_to_widget_coords(*self.start)
        end_x, end_y = image_to_widget_coords(*self.end)
        distance = math.hypot(end_x - start_x, end_y - start_y)
        if distance < self.MIN_DISTANCE_THRESHOLD:
            return

        width = self.options.size * 1.75
        arrow_head_size = self.options.size * self.ARROW_HEAD_SIZE_MULTIPLIER * 1.75

        arrow_scale = min(1.0, max(self.MIN_ARROW_SCALE, distance / (self.SCALE_DISTANCE_THRESHOLD * scale)))
        scaled_arrow_head_size = arrow_head_size * arrow_scale

        angle = math.atan2(end_y - start_y, end_x - start_x)

        cr.set_line_width(width * scale)
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        cr.set_source_rgba(*self.options.primary_color)

        cr.move_to(start_x, start_y)
        cr.line_to(end_x, end_y)
        cr.stroke()

        head_length = scaled_arrow_head_size * scale

        left_angle = angle + math.pi - self.ARROW_ANGLE
        right_angle = angle + math.pi + self.ARROW_ANGLE

        left_x = end_x + head_length * math.cos(left_angle)
        left_y = end_y + head_length * math.sin(left_angle)

        right_x = end_x + head_length * math.cos(right_angle)
        right_y = end_y + head_length * math.sin(right_angle)

        cr.move_to(end_x, end_y)
        cr.line_to(left_x, left_y)
        cr.stroke()

        cr.move_to(end_x, end_y)
        cr.line_to(right_x, right_y)
        cr.stroke()

    def get_bounds(self) -> QuadBounds:
        distance = math.hypot(self.end[0] - self.start[0], self.end[1] - self.start[1])
        if distance < self.MIN_DISTANCE_THRESHOLD:
            return QuadBounds.from_rect(self.start[0], self.start[1], self.start[0], self.start[1])

        arrow_head_size = self.options.size * self.ARROW_HEAD_SIZE_MULTIPLIER * 1.75
        arrow_scale = min(1.0, max(self.MIN_ARROW_SCALE, distance / self.SCALE_DISTANCE_THRESHOLD))
        scaled_arrow_head_size = arrow_head_size * arrow_scale

        angle = math.atan2(self.end[1] - self.start[1], self.end[0] - self.start[0])
        perp_cos = -math.sin(angle)
        perp_sin = math.cos(angle)

        start_x, start_y = self.start
        end_x, end_y = self.end

        max_extent = scaled_arrow_head_size

        p1 = (start_x + max_extent * perp_cos, start_y + max_extent * perp_sin)
        p2 = (end_x + max_extent * perp_cos, end_y + max_extent * perp_sin)
        p3 = (end_x - max_extent * perp_cos, end_y - max_extent * perp_sin)
        p4 = (start_x - max_extent * perp_cos, start_y - max_extent * perp_sin)

        return QuadBounds(p1, p2, p3, p4)

    def translate(self, dx: int, dy: int):
        self.start = (self.start[0] + dx, self.start[1] + dy)
        self.end = (self.end[0] + dx, self.end[1] + dy)

class TextAction(DrawingAction):
    PADDING_X_IMG = 4
    PADDING_Y_IMG = 2

    def __init__(self, position: tuple[int, int], text: str, intrinsic_image_bounds: tuple[int, int], options, font_size):
        self.options = options
        self.position = position
        self.text = text
        self.intrinsic_image_bounds = intrinsic_image_bounds
        self.font_size = font_size

    def contains_emoji(self) -> bool:
        for char in self.text:
            cat = unicodedata.category(char)
            if cat.startswith("S") or ord(char) > 0xFFFF:
                return True
            if "EMOJI" in unicodedata.name(char, "").upper():
                return True
        return False

    def draw_rounded_rectangle(self, cr: cairo.Context, x: float, y: float, width: float, height: float, radius: float, round_top: bool = True, round_bottom: bool = True):
        cr.new_sub_path()
        if round_top:
            cr.arc(x + radius, y + radius, radius, math.pi, 3 * math.pi / 2)
            cr.arc(x + width - radius, y + radius, radius, 3 * math.pi / 2, 0)
        else:
            cr.move_to(x, y)
            cr.line_to(x + width, y)

        if round_bottom:
            cr.arc(x + width - radius, y + height - radius, radius, 0, math.pi / 2)
            cr.arc(x + radius, y + height - radius, radius, math.pi / 2, math.pi)
        else:
            cr.line_to(x + width, y + height)
            cr.line_to(x, y + height)

        cr.close_path()

    def draw_per_line_background(self, cr: cairo.Context, layout, text_x_widget: float, text_y_widget: float, scale: float):
        lines = self.text.split('\n')
        if len(lines) <= 1:
            _, logical_rect = layout.get_extents()
            text_width_widget = logical_rect.width / Pango.SCALE
            text_height_widget = logical_rect.height / Pango.SCALE

            bg_x_widget = text_x_widget - self.PADDING_X_IMG * scale
            bg_y_widget = text_y_widget - self.PADDING_Y_IMG * scale
            bg_width_widget = text_width_widget + 2 * self.PADDING_X_IMG * scale
            bg_height_widget = text_height_widget + 2 * self.PADDING_Y_IMG * scale

            radius = min(6.0 * scale, min(bg_width_widget, bg_height_widget) / 4)
            self.draw_rounded_rectangle(cr, bg_x_widget, bg_y_widget, bg_width_widget, bg_height_widget, radius)
            cr.fill()
            return

        line_heights = []
        line_widths = []

        _, overall_logical_rect = layout.get_extents()
        overall_width = overall_logical_rect.width / Pango.SCALE

        for line in lines:
            temp_layout = PangoCairo.create_layout(cr)
            font_desc = Pango.FontDescription()
            font_desc.set_family(self.options.font)
            font_desc.set_size(int(self.font_size * scale * Pango.SCALE))
            temp_layout.set_font_description(font_desc)
            temp_layout.set_text(line, -1)
            temp_layout.set_alignment(Pango.Alignment.CENTER)

            _, logical_rect = temp_layout.get_extents()
            line_widths.append(logical_rect.width / Pango.SCALE)
            line_heights.append(logical_rect.height / Pango.SCALE)

        current_y = text_y_widget

        for i, (line_width, line_height) in enumerate(zip(line_widths, line_heights)):
            bg_x_widget = text_x_widget + (overall_width - line_width) / 2 - self.PADDING_X_IMG * scale
            bg_y_widget = current_y - self.PADDING_Y_IMG * scale
            bg_width_widget = line_width + 2 * self.PADDING_X_IMG * scale
            bg_height_widget = line_height + 2 * self.PADDING_Y_IMG * scale

            radius = min(6.0 * scale, min(bg_width_widget, bg_height_widget) / 4)

            round_top = True
            round_bottom = True

            if i > 0 and line_widths[i-1] >= line_width:
                round_top = False
            if i < len(lines) - 1 and line_widths[i+1] >= line_width:
                round_bottom = False

            self.draw_rounded_rectangle(cr, bg_x_widget, bg_y_widget, bg_width_widget, bg_height_widget, radius, round_top, round_bottom)
            cr.fill()

            current_y += line_height

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        if not self.text.strip():
            return

        x_widget, y_widget = image_to_widget_coords(*self.position)

        layout = PangoCairo.create_layout(cr)
        font_desc = Pango.FontDescription()
        font_desc.set_family(self.options.font)
        font_desc.set_size(int(self.font_size * scale * Pango.SCALE))
        layout.set_font_description(font_desc)
        layout.set_text(self.text, -1)
        layout.set_alignment(Pango.Alignment.CENTER)

        _, logical_rect = layout.get_extents()
        text_width_widget = logical_rect.width / Pango.SCALE
        text_height_widget = logical_rect.height / Pango.SCALE

        text_x_widget = x_widget - text_width_widget / 2
        text_y_widget = y_widget - text_height_widget

        if self.options.fill_color and any(c > 0 for c in self.options.fill_color):
            cr.set_source_rgba(*self.options.fill_color)
            self.draw_per_line_background(cr, layout, text_x_widget, text_y_widget, scale)

        cr.move_to(text_x_widget, text_y_widget)
        if self.contains_emoji():
            cr.set_source_rgba(*self.options.primary_color)
            PangoCairo.show_layout(cr, layout)
        else:
            PangoCairo.layout_path(cr, layout)
            if self.options.border_color and any(c > 0 for c in self.options.border_color):
                cr.set_source_rgba(*self.options.border_color)
                base_line_width = 2.0
                adjusted_line_width = base_line_width * scale * (self.font_size / 14.0)
                cr.set_line_width(adjusted_line_width)
                cr.stroke_preserve()
            cr.set_source_rgba(*self.options.primary_color)
            cr.fill()

    def get_bounds(self) -> QuadBounds:
        if not self.text.strip():
            x, y = self.position
            return QuadBounds.from_rect(x, y, x, y)

        temp_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1, 1)
        temp_cr = cairo.Context(temp_surface)
        layout = PangoCairo.create_layout(temp_cr)
        font_desc = Pango.FontDescription()
        font_desc.set_family(self.options.font)
        font_desc.set_size(int(self.font_size * Pango.SCALE))
        layout.set_font_description(font_desc)
        layout.set_text(self.text, -1)
        layout.set_alignment(Pango.Alignment.CENTER)

        _, logical_rect = layout.get_extents()
        text_width_img = int(logical_rect.width / Pango.SCALE)
        text_height_img = int(logical_rect.height / Pango.SCALE)

        x_img, y_img = self.position

        outline_padding = 0
        if self.options.border_color and any(c > 0 for c in self.options.border_color):
            outline_padding = int(2.0 * (self.font_size / 14.0)) + 1

        left_img = x_img - text_width_img // 2 - self.PADDING_X_IMG - outline_padding
        right_img = x_img + text_width_img // 2 + self.PADDING_X_IMG + outline_padding
        top_img = y_img - text_height_img - self.PADDING_Y_IMG - outline_padding
        bottom_img = y_img + self.PADDING_Y_IMG + outline_padding

        return QuadBounds.from_rect(left_img, top_img, right_img, bottom_img)

    def translate(self, dx: int, dy: int):
        self.position = (self.position[0] + dx, self.position[1] + dy)


class LineAction(ArrowAction):
    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        start_x, start_y = image_to_widget_coords(*self.start)
        end_x, end_y = image_to_widget_coords(*self.end)

        width = self.options.size * 1.75
        angle = math.atan2(end_y - start_y, end_x - start_x)
        half_width = (width * scale) / 2

        start_x += half_width * math.cos(angle)
        start_y += half_width * math.sin(angle)
        end_x -= half_width * math.cos(angle)
        end_y -= half_width * math.sin(angle)

        line_width = width * scale
        cr.set_line_width(line_width)
        cr.move_to(start_x, start_y)
        cr.line_to(end_x, end_y)
        cr.set_source_rgba(*self.options.primary_color)
        cr.stroke()

    def get_bounds(self) -> QuadBounds:
        angle = math.atan2(self.end[1] - self.start[1], self.end[0] - self.start[0])

        width = self.options.size * 1.75
        half_width = width / 2

        cos_angle = math.cos(angle)
        sin_angle = math.sin(angle)
        perp_cos = -sin_angle
        perp_sin = cos_angle

        start_x, start_y = self.start
        end_x, end_y = self.end

        p1 = (start_x + half_width * perp_cos, start_y + half_width * perp_sin)
        p2 = (end_x + half_width * perp_cos, end_y + half_width * perp_sin)
        p3 = (end_x - half_width * perp_cos, end_y - half_width * perp_sin)
        p4 = (start_x - half_width * perp_cos, start_y - half_width * perp_sin)

        return QuadBounds(p1, p2, p3, p4)


class RectAction(DrawingAction):
    def __init__(self, start: tuple[int, int], end: tuple[int, int], shift: bool, options):
        self.options = options
        self.start = start
        self.end = end
        self.shift = shift

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        x1_widget, y1_widget = image_to_widget_coords(*self.start)
        x2_widget, y2_widget = image_to_widget_coords(*self.end)

        if self.shift:
            size = max(abs(x2_widget - x1_widget), abs(y2_widget - y1_widget))
            if x2_widget < x1_widget:
                x2_widget = x1_widget - size
            else:
                x2_widget = x1_widget + size
            if y2_widget < y1_widget:
                y2_widget = y1_widget - size
            else:
                y2_widget = y1_widget + size

        stroke_offset = (self.options.size * scale) / 2
        x = min(x1_widget, x2_widget) + stroke_offset
        y = min(y1_widget, y2_widget) + stroke_offset
        w = abs(x2_widget - x1_widget) - (self.options.size * scale)
        h = abs(y2_widget - y1_widget) - (self.options.size * scale)

        if w > 0 and h > 0:
            if self.options.fill_color:
                cr.set_source_rgba(*self.options.fill_color)
                cr.rectangle(x, y, w, h)
                cr.fill()
            cr.set_source_rgba(*self.options.primary_color)
            cr.set_line_width(self.options.size * scale)
            cr.rectangle(x, y, w, h)
            cr.stroke()

    def get_bounds(self) -> QuadBounds:
        if self.shift:
            dx = abs(self.end[0] - self.start[0])
            dy = abs(self.end[1] - self.start[1])
            size = max(dx, dy)
            if self.end[0] < self.start[0]:
                end_x = self.start[0] - size
            else:
                end_x = self.start[0] + size
            if self.end[1] < self.start[1]:
                end_y = self.start[1] - size
            else:
                end_y = self.start[1] + size
            return QuadBounds.from_start_end(self.start, (end_x, end_y))
        else:
            return QuadBounds.from_start_end(self.start, self.end)

    def translate(self, dx: int, dy: int):
        self.start = (self.start[0] + dx, self.start[1] + dy)
        self.end = (self.end[0] + dx, self.end[1] + dy)


class CircleAction(RectAction):
    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        x1_widget, y1_widget = image_to_widget_coords(*self.start)
        x2_widget, y2_widget = image_to_widget_coords(*self.end)

        if self.shift:
            size = max(abs(x2_widget - x1_widget), abs(y2_widget - y1_widget))
            if x2_widget < x1_widget:
                x2_widget = x1_widget - size
            else:
                x2_widget = x1_widget + size
            if y2_widget < y1_widget:
                y2_widget = y1_widget - size
            else:
                y2_widget = y1_widget + size

        cx, cy = (x1_widget + x2_widget) / 2, (y1_widget + y2_widget) / 2
        stroke_offset = (self.options.size * scale) / 2
        rx = (abs(x2_widget - x1_widget) - (self.options.size * scale)) / 2
        ry = (abs(y2_widget - y1_widget) - (self.options.size * scale)) / 2

        if rx > 0 and ry > 0:
            cr.save()
            cr.translate(cx, cy)
            cr.scale(rx, ry)
            cr.arc(0, 0, 1, 0, 2 * math.pi)
            cr.restore()

            if self.options.fill_color:
                cr.set_source_rgba(*self.options.fill_color)
                cr.fill_preserve()
            cr.set_source_rgba(*self.options.primary_color)
            cr.set_line_width(self.options.size * scale)
            cr.stroke()


class HighlighterAction(StrokeAction):
    def __init__(self, stroke: list[tuple[int, int]], options, shift: bool):
        if shift and len(stroke) >= 2:
            start_point = stroke[0]
            end_point = stroke[-1]
            self.stroke = [start_point, (end_point[0], start_point[1])]
        else:
            self.stroke = stroke
        self.options = options

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        if len(self.stroke) < 2:
            return
        coords = [image_to_widget_coords(x, y) for x, y in self.stroke]
        cr.set_operator(cairo.Operator.MULTIPLY)
        cr.set_source_rgba(*self.options.primary_color)
        cr.set_line_width(self.options.size * scale * 2)
        cr.set_line_cap(cairo.LineCap.BUTT)
        cr.move_to(*coords[0])
        for point in coords[1:]:
            cr.line_to(*point)
        cr.stroke()
        cr.set_operator(cairo.Operator.OVER)
        cr.set_line_cap(cairo.LineCap.ROUND)

    def get_bounds(self) -> QuadBounds:
        if not self.stroke:
            return QuadBounds.from_rect(0, 0, 0, 0)
        xs, ys = zip(*self.stroke)
        padding = self.options.size + 3
        return QuadBounds.from_rect(min(xs) - padding, min(ys) - padding, max(xs) + padding, max(ys) + padding)


class CensorAction(RectAction):
    def __init__(self, start: tuple[int, int], end: tuple[int, int], background_pixbuf: GdkPixbuf.Pixbuf, options):
        super().__init__(start, end, False, options)

        self.original_scale = 1.0
        self.base_block_size = 8
        self.background_pixbuf = background_pixbuf

    def set_original_scale(self, scale: float):
        self.original_scale = scale

    def _get_scaled_block_size(self, current_scale: float) -> float:
        scale_ratio = current_scale / self.original_scale
        return self.base_block_size * scale_ratio

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        crop = self._get_image_crop()
        if not crop:
            return

        img_w, img_h = self.background_pixbuf.get_width(), self.background_pixbuf.get_height()

        crop_x_img = crop['x'] - img_w / 2
        crop_y_img = crop['y'] - img_h / 2
        crop_x2_img = crop_x_img + crop['width']
        crop_y2_img = crop_y_img + crop['height']

        x1, y1 = image_to_widget_coords(int(crop_x_img), int(crop_y_img))
        x2, y2 = image_to_widget_coords(int(crop_x2_img), int(crop_y2_img))

        x, y = min(x1, x2), min(y1, y2)
        width, height = abs(x2 - x1), abs(y2 - y1)
        if width < 1 or height < 1:
            return

        scaled_block_size = self._get_scaled_block_size(scale)
        self._draw_pixelation(cr, crop, x, y, width, height, scaled_block_size)

    def _draw_pixelation(self, cr: cairo.Context, crop: dict, x: float, y: float, width: float, height: float, block_size: float):
        cr.save()
        cr.rectangle(x, y, width, height)
        cr.clip()

        blocks_x = max(1, int(width / block_size))
        blocks_y = max(1, int(height / block_size))

        tiny_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, blocks_x, blocks_y)
        tiny_cr = cairo.Context(tiny_surface)

        tiny_cr.set_operator(cairo.OPERATOR_CLEAR)
        tiny_cr.paint()
        tiny_cr.set_operator(cairo.OPERATOR_OVER)

        tiny_cr.scale(blocks_x / crop['width'], blocks_y / crop['height'])
        tiny_cr.translate(-crop['x'], -crop['y'])
        Gdk.cairo_set_source_pixbuf(tiny_cr, self.background_pixbuf, 0, 0)
        tiny_cr.paint()

        cr.translate(x, y)
        cr.scale(width / blocks_x, height / blocks_y)
        pattern = cairo.SurfacePattern(tiny_surface)
        pattern.set_filter(cairo.FILTER_NEAREST)
        cr.set_source(pattern)
        cr.paint()

        cr.restore()

    def _get_image_crop(self) -> dict | None:
        img_w, img_h = self.background_pixbuf.get_width(), self.background_pixbuf.get_height()
        x1 = int(self.start[0] + img_w / 2)
        y1 = int(self.start[1] + img_h / 2)
        x2 = int(self.end[0] + img_w / 2)
        y2 = int(self.end[1] + img_h / 2)
        x_start, x_end = sorted([max(0, min(x1, img_w)), max(0, min(x2, img_w))])
        y_start, y_end = sorted([max(0, min(y1, img_h)), max(0, min(y2, img_h))])
        width, height = x_end - x_start, y_end - y_start
        return {'x': x_start, 'y': y_start, 'width': width, 'height': height} if width > 0 and height > 0 else None

    def get_bounds(self) -> QuadBounds:
        return QuadBounds.from_start_end(self.start, self.end)


class NumberStampAction(DrawingAction):
    def __init__(self, position: tuple[int, int], number: int, options):
        super().__init__()
        self.options = options
        self.position = position
        self.number = number
        self.creation_time = time.time()

    def draw(self, cr: cairo.Context, image_to_widget_coords: Callable[[int, int], tuple[float, float]], scale: float):
        x_widget, y_widget = image_to_widget_coords(*self.position)
        r_widget = self.options.size * 2 * scale

        cr.set_source_rgba(*self.options.fill_color)
        cr.arc(x_widget, y_widget, r_widget, 0, 2 * math.pi)
        cr.fill_preserve()

        if self.options.border_color.alpha != 0 and self.options.fill_color.alpha != 0:
            cr.set_source_rgba(*self.options.border_color)
            cr.set_line_width(2.0 * scale)
            cr.stroke()
        else:
            cr.new_path()

        cr.select_font_face("Adwaita Sans", cairo.FontSlant.NORMAL, cairo.FontWeight.BOLD)
        cr.set_font_size(r_widget * 1.2)
        text = str(self.number)

        xbearing, ybearing, width, height, xadvance, yadvance = cr.text_extents(text)
        tx = x_widget - width / 2 - xbearing
        ty = y_widget + height / 2

        cr.move_to(tx, ty)
        cr.text_path(text)

        if self.options.border_color and any(c > 0 for c in self.options.border_color):
            cr.set_source_rgba(*self.options.border_color)
            cr.set_line_width(4 * scale)
            cr.stroke_preserve()

        cr.set_source_rgba(*self.options.primary_color)
        cr.fill()

    def contains_point(self, px_img: int, py_img: int) -> bool:
        x_img, y_img = self.position
        radius = self.options.size * 2
        distance_sq = (px_img - x_img)**2 + (py_img - y_img)**2
        return distance_sq <= (radius + 5)**2

    def get_bounds(self) -> QuadBounds:
        x_img, y_img = self.position
        outline_padding = 2 if self.options.border_color and any(c > 0 for c in self.options.border_color) else 0
        total_radius = self.options.size * 2 + outline_padding + 1
        return QuadBounds.from_rect(x_img - total_radius, y_img - total_radius, x_img + total_radius, y_img + total_radius)

    def translate(self, dx: int, dy: int):
        self.position = (self.position[0] + dx, self.position[1] + dy)



# Copyright (C) 2025 tfuxu, Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, GLib, Gdk, Gio, Gtk

from gradia.constants import rootdir  # pyright: ignore
from gradia.ui.recent_picker import RecentPicker
from gradia.overlay.drop_overlay import DropOverlay

@Gtk.Template(resource_path=f"{rootdir}/ui/welcome_page.ui")
class WelcomePage(Adw.BreakpointBin):
    __gtype_name__ = "GradiaWelcomePage"

    recent_picker: RecentPicker = Gtk.Template.Child()
    drop_overlay: DropOverlay = Gtk.Template.Child()

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.recent_picker.callback = self._on_recent_image_click
        self._setup_drag_and_drop()

    """
    Setup Methods
    """

    def _setup_drag_and_drop(self) -> None:
        drop_target = Gtk.DropTarget.new(Gio.File, Gdk.DragAction.COPY)
        drop_target.set_preload(True)

        drop_target.connect("drop", self._on_file_dropped)
        self.drop_overlay.drop_target = drop_target

    """
    Callbacks
    """

    def _on_file_dropped(self, _target: Gtk.DropTarget, value: Gio.File, _x: int, _y: int) -> bool:
        uri = value.get_uri()
        if uri:
            window = self.get_root()
            action = window.lookup_action("load-drop") if window else None
            if action:
                action.activate(GLib.Variant('s', uri))
                return True
        return False

    def refresh_recent_picker(self) -> None:
        self.recent_picker.refresh()

    """
    Callbacks
    """
    def _on_recent_image_click(self, path: str, gradient_index: int) -> None:
        window = self.get_root()
        if window:
            action = window.lookup_action("open-path")
            if action:
                param = GLib.Variant('s', path)
                action.activate(param)




# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gtk, GObject, Gdk, Gsk, Graphene, GLib
from gradia.ui.widget.angle_selector import AngleSelector
from gradia.constants import rootdir
from gradia.utils.colors import is_light_color_rgba
from typing import Optional, Callable, List, Tuple
import operator
import time
import math

class GradientColorButton(Gtk.Box):
    __gtype_name__ = "GradientColorButton"

    color = GObject.Property(
        type=Gdk.RGBA,
        default=Gdk.RGBA(),
        nick="Color",
    )

    step = GObject.Property(
        type=float,
        default=0.0,
        minimum=0.0,
        maximum=1.0,
        nick="Step",
    )

    selected = GObject.Property(
        type=bool,
        default=False,
        nick="Selected",
    )

    def __init__(self, tooltip_text: str = "", **kwargs):
        super().__init__(**kwargs)
        self.set_tooltip_text(tooltip_text)
        self.add_css_class("gradient-color-picker")

        self._css_provider = Gtk.CssProvider()
        self.get_style_context().add_provider(
            self._css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 2
        )

        box = Gtk.Box(halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER)
        self._icon = Gtk.Image.new_from_icon_name("edit-symbolic")
        self._icon.set_pixel_size(18)
        self._icon.set_size_request(30,-1)
        self._icon.get_style_context().add_class("gradient-color-button")
        box.append(self._icon)

        self.append(box)

        self._color_dialog: Optional[Gtk.ColorDialog] = None
        self.editor: Optional["GradientEditor"] = None
        self._has_moved = False

        self.connect("notify::color", self._on_color_changed)
        self.connect("notify::step", self._on_step_changed)
        self.connect("notify::selected", self._on_selected_changed)

        default_color = Gdk.RGBA()
        default_color.parse("#000000")
        self.set_property("color", default_color)

        self._update_color_css()

    def _on_color_changed(self, *_):
        self._update_color_css()

    def _on_step_changed(self, *_):
        self.set_tooltip_text(f"{scale_correction(self.step):.0%}")
        if self.editor:
            self.editor._update_ui_for_selected_button()

    def _on_selected_changed(self, *_):
        if self.selected:
            self.add_css_class("selected")
            if self.editor:
                for other in self.editor.color_buttons:
                    if other != self:
                        other.set_selected(False)
                self.editor._update_ui_for_selected_button()
        else:
            self.remove_css_class("selected")

    def _update_color_css(self):
        rgba = self.color
        color_str = rgba.to_string()
        css = f"""
            .gradient-color-picker {{
                background-color: {color_str};
                border-radius: 9999px;
            }}
        """
        self._css_provider.load_from_string(css)

        if is_light_color_rgba(rgba):
            self._icon.get_style_context().add_class("dark")
        else:
            self._icon.get_style_context().remove_class("dark")

    def set_step(self, step: float) -> None:
        self.step = max(0.0, min(1.0, step))

    def set_selected(self, selected: bool) -> None:
        self.selected = selected

    def set_color(self, color) -> None:
        if isinstance(color, str):
            rgba = Gdk.RGBA()
            rgba.parse(color)
            self.color = rgba
        else:
            self.color = color

    def get_color(self) -> Gdk.RGBA:
        return self.color

    def get_color_string(self) -> str:
        return self.color.to_string()

    def get_step(self) -> float:
        return self.step

    def get_selected(self) -> bool:
        return self.selected

    def open_color_picker(self, on_select: Optional[Callable[[], None]] = None) -> None:
        if self._color_dialog is None:
            self._color_dialog = Gtk.ColorDialog(with_alpha=False)

        root = self.get_root()
        current_color = self.color

        def on_color_chosen(dialog, result):
            try:
                new_color = dialog.choose_rgba_finish(result)
                self.set_property("color", new_color)

                if on_select:
                    on_select()
            except GLib.Error:
                pass

        self._color_dialog.choose_rgba(
            root,
            current_color,
            None,
            on_color_chosen
        )

class GradientEditor(Gtk.Box):
    __gtype_name__ = "GradientEditor"

    def __init__(self, **kwargs):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, **kwargs)

        self.set_hexpand(False)
        self.set_halign(Gtk.Align.CENTER)
        self.set_size_request(248, 34)
        self.add_css_class("gradient-editor")

        self.color_buttons: List[GradientColorButton] = []
        self.drag_offset: Optional[Tuple[float, float]] = None
        self.selector: Optional["GradientSelector"] = None
        self._gradient_changed_callback: Optional[Callable[[List[Tuple[float, str]]], None]] = None

        self.overlay = Gtk.Overlay()
        self.append(self.overlay)

        self.gradient_background = Gtk.Box()
        self.gradient_background.set_hexpand(False)
        self.gradient_background.set_size_request(248, 34)
        self.gradient_background.add_css_class("gradient-background")

        self.gradient_background.connect("notify::width", self._on_gradient_background_size_changed)

        self._gradient_css_provider = Gtk.CssProvider()
        self.gradient_background.get_style_context().add_provider(
            self._gradient_css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 1
        )

        self.button_container = Gtk.Fixed(overflow=Gtk.Overflow.VISIBLE)
        self.button_container.set_hexpand(False)
        self.button_container.set_size_request(248, 36)

        self.overlay.set_child(self.gradient_background)
        self.overlay.add_overlay(self.button_container)

        self.click_controller = Gtk.GestureClick()
        self.click_controller.connect("pressed", self._on_background_clicked)
        self.button_container.add_controller(self.click_controller)

    def set_on_gradient_changed(self, callback: Callable[[List[Tuple[float, str]]], None]):
        self._gradient_changed_callback = callback

    def _on_gradient_changed(self):
        if self._gradient_changed_callback:
            self._gradient_changed_callback(self.get_gradient_data())

    def _on_gradient_background_size_changed(self, widget, pspec):
        self._update_button_positions()

    def get_gradient_data(self) -> List[Tuple[float, str]]:
        sorted_buttons = sorted(self.color_buttons, key=lambda b: b.get_step())
        return [(btn.get_step(), btn.get_color_string()) for btn in sorted_buttons]


    def _add_color_button(self, button: GradientColorButton):
        self.color_buttons.append(button)
        button.editor = self

        button.connect("notify::color", self._on_color_changed)

        drag_controller = Gtk.GestureDrag()
        drag_controller.set_touch_only(False)
        drag_controller.connect("drag-begin", self._on_drag_begin)
        drag_controller.connect("drag-update", self._on_drag_update)
        drag_controller.connect("drag-end", self._on_drag_end)
        button.add_controller(drag_controller)

        self.button_container.put(button, 0, -2)
        self._on_gradient_changed()

    def _remove_button(self, button: GradientColorButton):
        if len(self.color_buttons) > 2:
            was_selected = button.get_selected()
            self.color_buttons.remove(button)
            self.button_container.remove(button)
            button.editor = None
            self._update_ui_for_selected_button()

            self._update_gradient_css()
            self._on_gradient_changed()

    def _get_container_pos(self, button):
        pointer = self.get_display().get_default_seat().get_pointer()
        _, px, py, _ = self.button_container.get_native().get_surface().get_device_position(pointer)
        return self.translate_coordinates(self.button_container, px, py)

    def _on_drag_begin(self, controller, start_x, start_y):
        button = controller.get_widget()
        button._has_moved = False
        button.set_selected(True)
        current_pos = self.button_container.get_child_position(button)
        container_pos = self._get_container_pos(button)
        self.drag_offset = (current_pos.x - container_pos[0], current_pos.y - container_pos[1])

    def _on_drag_update(self, controller, offset_x, offset_y):
        if self.drag_offset is None:
            return

        button = controller.get_widget()

        drag_distance = math.sqrt(offset_x * offset_x + offset_y * offset_y)
        if drag_distance > 1:
            button._has_moved = True

        container_width = self.gradient_background.get_allocated_width()
        button_width = button.get_allocated_width()

        if container_width <= 0 or button_width <= 0:
            return

        pointer_x_in_container, _ = self._get_container_pos(button)

        new_x = pointer_x_in_container + self.drag_offset[0]

        min_x = 0
        max_x = container_width - button_width
        clamped_x = max(min_x, min(max_x, new_x))

        button_center_x = clamped_x + (button_width / 2.0)
        new_step = button_center_x / container_width

        new_step = max(0.00, min(1.00, new_step))

        if not self._check_overlap_at_step(button, new_step):
            button.set_step(new_step)
            step_position = new_step * container_width - (button_width / 2.0)
            step_position = max(0, min(container_width - button_width, step_position))
            self.button_container.move(button, int(step_position), 0)
            self._update_gradient_css()

    def _on_drag_end(self, controller, offset_x, offset_y):
        button = controller.get_widget()

        if not button._has_moved:
            button.open_color_picker(on_select=lambda: self._on_gradient_changed())
        else:
            self._on_gradient_changed()

        self.drag_offset = None
        self._update_gradient_css()

    def _check_overlap_at_step(self, moving_button: GradientColorButton, step: float) -> bool:
        container_width = self.gradient_background.get_allocated_width()
        button_width = moving_button.get_allocated_width()
        if button_width <= 0:
            button_width = 34

        proposed_center = step * container_width
        proposed_left = proposed_center - (button_width / 2)
        proposed_right = proposed_left + button_width

        overlap_threshold = 0

        for other_button in self.color_buttons:
            if other_button == moving_button:
                continue

            other_step = other_button.get_step()
            other_width = other_button.get_allocated_width()
            if other_width <= 0:
                other_width = 34

            other_center = other_step * container_width
            other_left = other_center - (other_width / 2)
            other_right = other_left + other_width

            if not (proposed_right <= other_left + overlap_threshold or
                    other_right <= proposed_left + overlap_threshold):
                return True

        return False

    def _on_background_clicked(self, gesture, n_press, x, y):
        if n_press == 1:
            width = self.gradient_background.get_allocated_width()

            if len(self.color_buttons) >= 5:
                return;

            step = (x / width)
            step = max(0.0, min(1.0, step))

            min_distance = 0.13
            is_near_existing = any(abs(btn.get_step() - step) < min_distance for btn in self.color_buttons)

            if not is_near_existing:
                color = self._interpolate_color_at_step(step)
                self._create_button_at_step(step, color)

    def _create_button_at_step(self, step: float, color: Gdk.RGBA):
        button = GradientColorButton()
        button.set_step(step)
        button.set_color(color)
        button.add_css_class("gradient-editor-button")

        self._add_color_button(button)
        button.set_selected(True)
        self.color_buttons.sort(key=lambda b: b.get_step())
        self._update_button_positions()
        self._update_gradient_css()

    def _interpolate_color_at_step(self, step: float) -> Gdk.RGBA:
        sorted_buttons = sorted(self.color_buttons, key=lambda b: b.get_step())

        if step <= sorted_buttons[0].get_step():
            return sorted_buttons[0].get_color()
        if step >= sorted_buttons[-1].get_step():
            return sorted_buttons[-1].get_color()

        for i in range(len(sorted_buttons) - 1):
            btn1, btn2 = sorted_buttons[i], sorted_buttons[i + 1]
            if btn1.get_step() <= step <= btn2.get_step():
                if (btn2.get_step() - btn1.get_step()) == 0:
                    return btn1.get_color()
                t = (step - btn1.get_step()) / (btn2.get_step() - btn1.get_step())
                result = Gdk.RGBA()
                result.red = btn1.get_color().red + t * (btn2.get_color().red - btn1.get_color().red)
                result.green = btn1.get_color().green + t * (btn2.get_color().green - btn1.get_color().green)
                result.blue = btn1.get_color().blue + t * (btn2.get_color().blue - btn1.get_color().blue)
                result.alpha = btn1.get_color().alpha + t * (btn2.get_color().alpha - btn1.get_color().alpha)
                return result

        return sorted_buttons[0].get_color()

    def _on_color_changed(self, button, pspec):
        self._update_gradient_css()

    def _update_button_positions(self):
        width = 248

        for button in self.color_buttons:
            step = button.get_step()
            button_width = 34
            center_x = step * width
            x = center_x - (button_width / 2.0)
            x = max(0, min(width - button_width, x))

            x = int(x)
            self.button_container.move(button, x, 0)

        self.button_container.queue_draw()

    def _update_gradient_css(self):
        if not self.color_buttons:
            return

        self.color_buttons.sort(key=lambda b: b.get_step())
        sorted_buttons = self.color_buttons

        if len(sorted_buttons) == 1:
            color = sorted_buttons[0].get_color_string()
            css = f"""
                .gradient-background {{
                    background: {color};
                }}
            """
        else:
            stops = []
            for button in sorted_buttons:
                color = button.get_color_string()
                step = button.get_step()
                stops.append(f"{color} {step * 100}%")

            gradient_stops = ", ".join(stops)
            css = f"""
                .gradient-background {{
                    background: linear-gradient(to right, {gradient_stops});
                }}
            """
        self._gradient_css_provider.load_from_string(css)

    def _update_ui_for_selected_button(self):
        if not self.selector:
            return

        selected_button = None
        for button in self.color_buttons:
            if button.get_selected():
                selected_button = button
                break

        if selected_button:
            step = selected_button.get_step()
            self.selector.step_label.set_label(f"{scale_correction(step):.0%}")
            self.selector.button_revealer.set_reveal_child(True)
            self.selector.remove_button_revealer.set_reveal_child(len(self.color_buttons) > 2)
        else:
            self.selector.button_revealer.set_reveal_child(False)

    def get_selected_button(self) -> Optional[GradientColorButton]:
        for button in self.color_buttons:
            if button.get_selected():
                return button
        return None

    def remove_selected_button(self):
        selected_button = self.get_selected_button()
        if selected_button:
            self._remove_button(selected_button)

    def get_gradient_data(self) -> List[Tuple[float, str]]:
        sorted_buttons = sorted(self.color_buttons, key=lambda b: b.get_step())
        return [(scale_correction(btn.get_step()), btn.get_color_string()) for btn in sorted_buttons]

    def set_gradient_data(self, data: List[Tuple[float, str]]):
        for button in self.color_buttons[:]:
            self.button_container.remove(button)
        self.color_buttons.clear()

        for i, (step, color_string) in enumerate(data):
            step = reverse_scale_correction(step)
            button = GradientColorButton()
            button.set_step(step)
            button.set_color(color_string)
            button.add_css_class("gradient-editor-button")
            self._add_color_button(button)

        self._update_button_positions()
        self._update_gradient_css()

def scale_correction(value: float) -> float:
    return (value - 0.07) / (0.93 - 0.07)

def reverse_scale_correction(value: float) -> float:
    return value * (0.93 - 0.07) + 0.07



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk, Gdk, Graphene, Gsk, GObject, Adw
import math

class AngleSelector(Gtk.Widget):
    __gtype_name__ = 'GradiaAngleSelector'

    angle = GObject.Property(
        type=int,
        default=0,
        minimum=0,
        maximum=360,
        nick='Angle (degrees)',
        blurb='Angle in degrees between 0 and 360'
    )

    __gsignals__ = {
        'angle-changed': (GObject.SignalFlags.RUN_FIRST, None, (int,))
    }

    def __init__(self):
        super().__init__(css_name='angle-selector')
        self._outer_radius = 40
        self._inner_radius = 37
        self._handle_size = 20
        self._handle_radius = 10
        self._dragging = False
        self._snap_threshold = 15
        self._exit_threshold = 25

        self.set_size_request(100, 100)
        self.set_can_focus(True)
        self._setup_gestures()
        self.connect("notify::angle", lambda *a: self.queue_draw())

    def do_get_css_name(self):
        return "angle-selector"

    def _setup_gestures(self):
        drag_controller = Gtk.GestureDrag.new()
        drag_controller.connect('drag-begin', self._on_drag_begin)
        drag_controller.connect('drag-update', self._on_drag_update)
        drag_controller.connect('drag-end', self._on_drag_end)
        self.add_controller(drag_controller)

        click_controller = Gtk.GestureClick.new()
        click_controller.connect('pressed', self._on_click)
        self.add_controller(click_controller)

    def _get_center(self):
        allocation = self.get_allocation()
        return allocation.width / 2, allocation.height / 2

    def _get_handle_position(self):
        cx, cy = self._get_center()
        adjusted_angle = self.angle - 90
        rad = math.radians(adjusted_angle)
        x = cx + self._inner_radius * math.cos(rad)
        y = cy + self._inner_radius * math.sin(rad)
        return x, y

    def _point_to_angle(self, x, y):
        cx, cy = self._get_center()
        dx = x - cx
        dy = y - cy
        angle_rad = math.atan2(dy, dx)
        angle_degrees = math.degrees(angle_rad)
        angle_degrees = (angle_degrees + 90) % 360
        if angle_degrees < 0:
            angle_degrees += 360
        return angle_degrees

    def _snap_to_90_degrees(self, angle):
        snap_angles = [0, 90, 180, 270]
        for snap_angle in snap_angles:
            if self.angle == snap_angle:
                distance = abs(angle - snap_angle)
                if distance > 180:
                    distance = 360 - distance
                if distance <= self._exit_threshold:
                    return snap_angle
                else:
                    return angle
        return angle

    def _is_point_in_handle(self, x, y):
        hx, hy = self._get_handle_position()
        dx = x - hx
        dy = y - hy
        distance = math.sqrt(dx * dx + dy * dy)
        return distance <= self._handle_radius + 5

    def _on_click(self, gesture, n_press, x, y):
        if self._is_point_in_handle(x, y):
            return
        new_angle = self._point_to_angle(x, y)
        snapped_angle = self._snap_to_90_degrees(new_angle)
        self.set_angle_internal(snapped_angle)
        self.emit('angle-changed', self.angle)

    def _on_drag_begin(self, gesture, start_x, start_y):
        if self._is_point_in_handle(start_x, start_y):
            self._dragging = True
            self.grab_focus()
        else:
            new_angle = self._point_to_angle(start_x, start_y)
            snapped_angle = self._snap_to_90_degrees(new_angle)
            self.set_angle_internal(snapped_angle)
            self._dragging = True
            self.grab_focus()

    def _on_drag_update(self, gesture, offset_x, offset_y):
        if not self._dragging:
            return
        point = gesture.get_start_point()
        current_x = point.x + offset_x
        current_y = point.y + offset_y
        new_angle = self._point_to_angle(current_x, current_y)
        snapped_angle = self._snap_to_90_degrees(new_angle)
        self.set_angle_internal(snapped_angle)

    def _on_drag_end(self, gesture, offset_x, offset_y):
        self._dragging = False
        self.emit('angle-changed', self.angle)

    def set_angle_internal(self, degrees):
        degrees = degrees % 360
        if abs(self.angle - degrees) > 0.5:
            self.angle = int(round(degrees))
            self.queue_draw()

    def get_angle_radians(self):
        return math.radians(self.angle)

    def set_angle_radians(self, radians_val):
        degrees = math.degrees(radians_val) % 360
        self.set_angle_internal(degrees)

    def do_snapshot(self, snapshot):
        allocation = self.get_allocation()
        width, height = allocation.width, allocation.height
        cx, cy = width / 2, height / 2
        context = self.get_style_context()
        bg_color = context.get_color()
        fg_color = context.get_color()
        luminance = 0.299 * bg_color.red + 0.587 * bg_color.green + 0.114 * bg_color.blue
        is_dark = luminance < 0.5

        if not is_dark:
            background_color = Gdk.RGBA(0.32,0.32,0.33,1)
            foreground_color = Gdk.RGBA(0.82,0.82,0.82,1)
            shadow_color = Gdk.RGBA(0,0,0.023,1)
        else:
            background_color = Gdk.RGBA(0.88,0.88,0.88,1)
            foreground_color = Gdk.RGBA(1,1,1,1)
            shadow_color = Gdk.RGBA(0,0,0.023,1)

        self._draw_circle_border(snapshot, cx, cy, self._outer_radius, background_color)
        self._draw_90_degree_lines(snapshot, cx, cy, fg_color)
        self._draw_handle(snapshot, cx, cy, foreground_color, shadow_color)

    def _draw_90_degree_lines(self, snapshot, cx, cy, color):
        line_length = 4
        outer_start = self._outer_radius - 8
        inner_end = outer_start - line_length
        thickness = 1
        angles = [0, 90, 180, 270]
        for angle in angles:
            rad = math.radians(angle - 90)
            start_x = cx + outer_start * math.cos(rad)
            start_y = cy + outer_start * math.sin(rad)
            end_x = cx + inner_end * math.cos(rad)
            end_y = cy + inner_end * math.sin(rad)
            if abs(start_x - end_x) > abs(start_y - end_y):
                line_rect = Graphene.Rect()
                line_rect.init(min(start_x, end_x), cy - thickness/2, abs(end_x - start_x), thickness)
                snapshot.append_color(color, line_rect)
            else:
                line_rect = Graphene.Rect()
                line_rect.init(cx - thickness/2, min(start_y, end_y), thickness, abs(end_y - start_y))
                snapshot.append_color(color, line_rect)

    def _draw_handle(self, snapshot, cx, cy, color ,shadow_color):
        hx, hy = self._get_handle_position()
        angle_to_center = math.atan2(cy - hy, cx - hx)
        rotation_degrees = math.degrees(angle_to_center) - 45
        handle_color = color
        shadow_color1 = Gdk.RGBA(red=shadow_color.red, green=shadow_color.green, blue=shadow_color.blue, alpha=0.1)
        shadow_color2 = Gdk.RGBA(red=shadow_color.red, green=shadow_color.green, blue=shadow_color.blue, alpha=0.2)
        snapshot.save()
        snapshot.translate(Graphene.Point().init(hx, hy))
        snapshot.rotate(rotation_degrees)
        snapshot.translate(Graphene.Point().init(-self._handle_radius, -self._handle_radius))
        handle_rect = Graphene.Rect()
        handle_rect.init(0, 0, self._handle_size, self._handle_size)
        rounded_rect = Gsk.RoundedRect()
        rounded_rect.init(
            handle_rect,
            Graphene.Size().init(self._handle_radius, self._handle_radius),
            Graphene.Size().init(self._handle_radius, self._handle_radius),
            Graphene.Size().init(0, 0),
            Graphene.Size().init(self._handle_radius, self._handle_radius)
        )
        shadow_node1 = Gsk.OutsetShadowNode.new(
            rounded_rect,
            shadow_color1,
            0, 0, 1, 0
        )
        shadow_node2 = Gsk.OutsetShadowNode.new(
            rounded_rect,
            shadow_color2,
            0, 0, 0, 4
        )
        snapshot.append_node(shadow_node1)
        snapshot.append_node(shadow_node2)
        snapshot.push_rounded_clip(rounded_rect)
        snapshot.append_color(handle_color, handle_rect)
        snapshot.pop()
        snapshot.restore()

    def _draw_circle_border(self, snapshot, cx, cy, radius, color):
        border_width = 4
        rect = Graphene.Rect()
        rect.init(cx - radius, cy - radius, radius * 2, radius * 2)
        rounded_rect = Gsk.RoundedRect()
        rounded_rect.init_from_rect(rect, radius)
        widths = [border_width] * 4
        colors = [color] * 4
        snapshot.append_node(
            Gsk.BorderNode.new(rounded_rect, widths, colors)
        )

AngleSelector.set_css_name("angle-selector")




# Copyright (C) 2026 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections.abc import Callable
from typing import Optional

from gi.repository import Adw, Gtk

from gradia.backend.ocr import OCR
from gradia.ui.dialog.ocr_dialog import OCRDialog
from gradia.ui.preferences.ocr_model_page import OCRModelPage


def present_ocr_dialog(
    image,
    parent: Optional[Gtk.Widget] = None,
    on_dialog_shown: Optional[Callable[[OCRDialog], None]] = None,
    on_cancelled: Optional[Callable[[], None]] = None,
    auto_copy: bool = False,
) -> None:
    def show():
        dialog = OCRDialog(image, auto_copy=auto_copy)
        dialog.present(parent)
        if on_dialog_shown is not None:
            on_dialog_shown(dialog)

    if len(OCR(parent).get_installed_models()) == 0:
        prefs = Adw.PreferencesDialog(content_width=600, content_height=500)
        prefs.set_title(_("OCR Language Models"))
        prefs.push_subpage(OCRModelPage(preferences_dialog=prefs, can_pop=False))

        def on_prefs_closed(*_):
            if len(OCR(parent).get_installed_models()) == 0:
                if on_cancelled is not None:
                    on_cancelled()
            else:
                show()

        prefs.connect("closed", on_prefs_closed)
        prefs.present(parent)
        return

    show()



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

rootdir          = '/be/alexandervanhee/gradia'
libdir           = '/usr/lib/gradia'
datadir          = '/usr/share'
pkgdatadir       = '/usr/share/gradia'
localedir        = '/usr/share/locale'

app_id           = 'be.alexandervanhee.gradia'
rel_ver          = '1.13.0'
version          = '1.13.0'
build_type       = 'plain'

project_url      = 'https://gradia.alexandervanhee.be/'
bugtracker_url   = 'https://github.com/AlexanderVanhee/Gradia/issues'
help_url         = 'https://github.com/AlexanderVanhee/Gradia/discussions'
#translate_url   = ''

# OCR paths
ocr_enabled            = 'True'
ocr_tesseract_cmd      = '/usr/bin/tesseract'
ocr_original_tessdata  = '/usr/share/tessdata'



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, Gdk, Gtk, GLib
from gradia.constants import app_id

class Settings:
    def __init__(self) -> None:
        self._settings = Gio.Settings.new(app_id)

    """
    Getters/Setters
    """

    @property
    def draw_mode(self) -> str:
        return self._settings.get_string("draw-mode")

    @draw_mode.setter
    def draw_mode(self, value: str) -> None:
        self._settings.set_string("draw-mode", value)

    def get_tool_config_item(self, key: str) -> str | None:
        tool_config = self.get_tool_config()
        return tool_config.get(key)

    def set_tool_config_item(self, key: str, value: str) -> None:
        tool_config = self.get_tool_config()
        tool_config[key] = value
        variant = GLib.Variant("a{ss}", tool_config)
        self._settings.set_value("tool-config", variant)

    def get_tool_config(self) -> dict[str, str]:
        variant = self._settings.get_value("tool-config")
        result = {}

        for i in range(variant.n_children()):
            entry = variant.get_child_value(i)

            key_variant = entry.get_child_value(0)
            value_variant = entry.get_child_value(1)

            key = key_variant.unpack()
            value = value_variant.unpack()

            result[key] = value

        return result


    @property
    def screenshot_folder(self) -> str:
        value = self._settings.get_string("screenshot-folder")
        return None if value == "" else value

    @screenshot_folder.setter
    def screenshot_folder(self, value: str) -> None:
        self._settings.set_string("screenshot-folder", value)

    @property
    def export_format(self) -> str:
        return self._settings.get_string("export-format")

    @export_format.setter
    def export_format(self, value: str) -> None:
        self._settings.set_string("export-format", value)

    @property
    def export_compress(self) -> bool:
        return self._settings.get_boolean("export-compress")

    @property
    def delete_screenshots_on_close(self) -> bool:
        return self._settings.get_boolean("trash-screenshots-on-close")

    @property
    def exit_method(self) -> str:
        value = self._settings.get_string("exit-method")
        if value in ("confirm", "copy", "none"):
            return value
        return "confirm"

    @exit_method.setter
    def exit_method(self, value: str) -> None:
        if value not in ("confirm", "copy", "none"):
            value = "confirm"
        self._settings.set_string("exit-method", value)

    @property
    def overwrite_screenshot(self) -> bool:
        return self._settings.get_boolean("overwrite-screenshot")

    @property
    def custom_export_command(self) -> str:
        return self._settings.get_string("custom-export-command")

    @custom_export_command.setter
    def custom_export_command(self, value: str) -> None:
        self._settings.set_string("custom-export-command", value)

    @property
    def provider_name(self) -> str:
        return self._settings.get_string("provider-name")

    @provider_name.setter
    def provider_name(self, value:str) -> None:
        self._settings.set_string("provider-name", value)

    @property
    def show_export_confirm_dialog(self) -> bool:
        return self._settings.get_boolean("show-export-confirm-dialog")

    @property
    def image_padding(self) -> int:
        return self._settings.get_int("image-padding")

    @image_padding.setter
    def image_padding(self, value: int) -> None:
        self._settings.set_int("image-padding", value)

    @property
    def image_corner_radius(self) -> int:
        return self._settings.get_int("image-corner-radius")

    @image_corner_radius.setter
    def image_corner_radius(self, value: int) -> None:
        self._settings.set_int("image-corner-radius", value)

    @property
    def image_aspect_ratio(self) -> str:
        return self._settings.get_string("image-aspect-ratio")

    @image_aspect_ratio.setter
    def image_aspect_ratio(self, value: str) -> None:
        self._settings.set_string("image-aspect-ratio", value)

    @property
    def image_shadow_strength(self) -> int:
        return self._settings.get_int("image-shadow-strength")

    @image_shadow_strength.setter
    def image_shadow_strength(self, value: int) -> None:
        self._settings.set_int("image-shadow-strength", value)

    @property
    def image_auto_balance(self) -> bool:
        return self._settings.get_boolean("image-auto-balance")

    @image_auto_balance.setter
    def image_auto_balance(self, value: bool) -> None:
        self._settings.set_boolean("image-auto-balance", value)

    @property
    def image_options_lock(self) -> bool:
        return self._settings.get_boolean("image-options-lock")

    @image_options_lock.setter
    def image_options_lock(self, value: bool) -> None:
        self._settings.set_boolean("image-options-lock", value)

    @property
    def background_mode(self) -> str:
        return self._settings.get_string("background-mode")

    @background_mode.setter
    def background_mode(self, value: str) -> None:
        self._settings.set_string("background-mode", value)

    @property
    def solid_state(self) -> str:
        return self._settings.get_string("solid-state")

    @solid_state.setter
    def solid_state(self, value: str) -> None:
        self._settings.set_string("solid-state", value)

    @property
    def gradient_state(self) -> str:
        return self._settings.get_string("gradient-state")

    @gradient_state.setter
    def gradient_state(self, value: str) -> None:
        self._settings.set_string("gradient-state", value)

    @property
    def image_state(self) -> str:
        return self._settings.get_string("image-state")

    @image_state.setter
    def image_state(self, value: str) -> None:
        self._settings.set_string("image-state", value)

    @property
    def source_snippet_language(self) -> str:
        return self._settings.get_string("source-snippet-language")

    @source_snippet_language.setter
    def source_snippet_language(self, value: str) -> None:
        self._settings.set_string("source-snippet-language", value)

    @property
    def source_snippet_code_text(self) -> str:
        return self._settings.get_string("source-snippet-code-text")

    @source_snippet_code_text.setter
    def source_snippet_code_text(self, value: str) -> None:
        self._settings.set_string("source-snippet-code-text", value)

    @property
    def source_snippet_title(self) -> str:
        return self._settings.get_string("source-snippet-title")

    @source_snippet_title.setter
    def source_snippet_title(self, value: str) -> None:
        self._settings.set_string("source-snippet-title", value)

    @property
    def source_snippet_style_scheme(self) -> str:
        return self._settings.get_string("source-snippet-style-scheme")

    @source_snippet_style_scheme.setter
    def source_snippet_style_scheme(self, value: str) -> None:
        self._settings.set_string("source-snippet-style-scheme", value)

    @property
    def trained_data(self) -> str:
        return self._settings.get_string("trained-data")

    @trained_data.setter
    def trained_data(self, value: str) -> None:
        self._settings.set_string("trained-data", value)

    """
    Internal Methods
    """

    def _parse_rgba(self, color_str: str, fallback: tuple[float, float, float, float]) -> Gdk.RGBA:
        rgba = Gdk.RGBA()

        try:
            parts = list(map(float, color_str.split(',')))

            if len(parts) == 4:
                rgba.red, rgba.green, rgba.blue, rgba.alpha = parts
            else:
                rgba.red, rgba.green, rgba.blue, rgba.alpha = fallback
        except (ValueError, IndexError):
            rgba.red, rgba.green, rgba.blue, rgba.alpha = fallback

        return rgba

    def _rgba_to_string(self, rgba: Gdk.RGBA) -> str:
        return f"{rgba.red:.3f},{rgba.green:.3f},{rgba.blue:.3f},{rgba.alpha:.3f}"

    def bind_switch(self, switch: Gtk.Switch, key: str) -> None:
        if key in self._settings.list_keys():
            self._settings.bind(
                key,
                switch,
                "active",
                Gio.SettingsBindFlags.DEFAULT
            )
        else:
            print(f"Warning: GSettings key '{key}' not found in schema.")

    def bind_adjustment(self, adjustment: Gtk.Adjustment, key: str) -> None:
        if key in self._settings.list_keys():
            self._settings.bind(
                key,
                adjustment,
                "value",
                Gio.SettingsBindFlags.DEFAULT
            )
        else:
            print(f"Warning: GSettings key '{key}' not found in schema.")

    def bind_scale(self, scale: Gtk.Scale, key: str) -> None:
        if key in self._settings.list_keys():
            self._settings.bind(
                key,
                scale.get_adjustment(),
                "value",
                Gio.SettingsBindFlags.DEFAULT
            )
        else:
            print(f"Warning: GSettings key '{key}' not found in schema.")

    def bind_spin_row(self, spin_row: object, key: str) -> None:
        if key in self._settings.list_keys():
            self._settings.bind(
                key,
                spin_row,
                "value",
                Gio.SettingsBindFlags.DEFAULT
            )
        else:
            print(f"Warning: GSettings key '{key}' not found in schema.")

    def bind_toggle_group(self, toggle_group, key: str):
        if key not in self._settings.list_keys():
            print(f"Warning: GSettings key '{key}' not found in schema.")
            return

        current = self._settings.get_string(key)
        toggle_group.set_active_name(current)

        def on_toggle_group_changed(toggle_group):
            self._settings.set_string(key, toggle_group.get_active_name())

        toggle_group.connect("notify::active-name", lambda w, p: on_toggle_group_changed(w))
